package com.minemaarten.signals.rail;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import net.minecraft.entity.item.EntityMinecart;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumFacing.Axis;
import net.minecraft.util.EnumFacing.AxisDirection;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.ChunkPos;
import net.minecraft.world.World;
import net.minecraft.world.chunk.Chunk;
import net.minecraftforge.event.world.BlockEvent.NeighborNotifyEvent;

import com.google.common.collect.Iterables;
import com.minemaarten.signals.lib.Log;
import com.minemaarten.signals.network.NetworkHandler;
import com.minemaarten.signals.network.PacketSyncStationNames;
import com.minemaarten.signals.tileentity.TileEntityStationMarker;

public class RailCacheManager{
    private static Map<Integer, RailCacheManager> INSTANCES = new HashMap<>();
    private final Map<Long, Map<BlockPos, RailWrapper>> railCache = new HashMap<>(); //Chunk hash to cache (per chunk) mapping.
    private final Set<TileEntityStationMarker> stations = new HashSet<>();
    private String[] allStationNames = new String[0];

    public static RailCacheManager getInstance(World world){
        if(world.field_72995_K) {
            Log.warning("Can't be called client side!");
            new Throwable().printStackTrace();
            return new RailCacheManager(); // Give back a clean cache client side
        }
        return getInstance(world.field_73011_w.getDimension());
    }

    public static RailCacheManager getInstance(int dimension){
        RailCacheManager cache = INSTANCES.get(dimension);
        if(cache == null) {
            cache = new RailCacheManager();
            INSTANCES.put(dimension, cache);
        }
        return cache;
    }

    public RailWrapper getRail(World world, BlockPos pos){
        return getRail(world, pos, true);
    }

    private RailWrapper getRail(World world, BlockPos pos, boolean shouldLoadRail){
        long chunkHash = getChunkHashFromBlockPos(pos);
        Map<BlockPos, RailWrapper> cache = railCache.get(chunkHash);
        if(cache != null) {
            RailWrapper rail = cache.get(pos);
            if(rail != null) return rail;
        }
        if(shouldLoadRail) {
            RailWrapper rail = new RailWrapper(world, pos);
            if(rail.isRail()) { //A rail isn't cached yet
                if(cache == null) {
                    cache = new HashMap<>();
                    railCache.put(chunkHash, cache);
                }
                cache.put(pos, rail);
                for(RailWrapper neighbor : rail.getNeighbors().keySet()) {
                    neighbor.updateNeighborCache();
                }
                NetworkController.getInstance(world).updateColor(rail, pos);
                return rail;
            }
        }
        return null;
    }

    private static long getChunkHashFromBlockPos(BlockPos pos){
        return ChunkPos.func_77272_a(pos.func_177958_n() >> 4, pos.func_177952_p() >> 4);
    }

    /**
     * When a chunk is loaded, we need to make sure to notify rails on a neighbor chunk, on a chunk border
     * @param chunk
     */
    public void onChunkLoad(Chunk chunk){
        for(EnumFacing d : EnumFacing.field_176754_o) {
            ChunkPos intPair = chunk.func_76632_l();
            BlockPos offsetBlock = intPair.func_180331_a(0, 0, 0).func_177967_a(d, 16);
            if(chunk.func_177412_p().func_175667_e(offsetBlock)) {
                Chunk neighbor = chunk.func_177412_p().func_72964_e(chunk.field_76635_g + d.func_82601_c(), chunk.field_76647_h + d.func_82599_e());
                Map<BlockPos, RailWrapper> chunkCache = railCache.get(ChunkPos.func_77272_a(neighbor.field_76635_g, neighbor.field_76647_h));
                if(chunkCache != null) {
                    if(d.func_176740_k() == Axis.X) {
                        int borderX = (d.func_176743_c() == AxisDirection.POSITIVE ? 0 : 15) + offsetBlock.func_177958_n();
                        for(RailWrapper rail : chunkCache.values()) {
                            if(rail.func_177958_n() == borderX) rail.updateNeighborCache();
                        }
                    } else {
                        int borderZ = (d.func_176743_c() == AxisDirection.POSITIVE ? 0 : 15) + offsetBlock.func_177952_p();
                        for(RailWrapper rail : chunkCache.values()) {
                            if(rail.func_177952_p() == borderZ) rail.updateNeighborCache();
                        }
                    }
                }
            }

        }
    }

    public void onChunkUnload(Chunk chunk){
        long chunkHash = ChunkPos.func_77272_a(chunk.field_76635_g, chunk.field_76647_h);
        Map<BlockPos, RailWrapper> cache = railCache.get(chunkHash);
        if(cache != null) {
            for(RailWrapper rail : cache.values()) {
                rail.invalidate();
            }
            railCache.remove(chunkHash);
        }
    }

    public void onWorldUnload(World world){
        INSTANCES.remove(world.field_73011_w.getDimension());
    }

    public void onNeighborChanged(NeighborNotifyEvent event){
        Map<BlockPos, RailWrapper> cache = railCache.get(getChunkHashFromBlockPos(event.getPos()));
        if(cache != null) {
            RailWrapper removedWrapper = cache.remove(event.getPos());
            if(removedWrapper != null) {
                removedWrapper.invalidate();
            }
        }
        RailWrapper rail = new RailWrapper(event.getWorld(), event.getPos(), event.getState());
        if(rail.isRail()) {
            for(EnumFacing d : EnumFacing.field_176754_o) {
                BlockPos neighborPos = event.getPos().func_177972_a(d);
                for(int i = -1; i <= 1; i++) {
                    RailWrapper neighbor = getRail(event.getWorld(), neighborPos.func_177982_a(0, i, 0), false);
                    if(neighbor != null) neighbor.updateNeighborCache();
                }
            }
        }
    }

    public void addStationMarker(TileEntityStationMarker marker){
        stations.add(marker);
        NetworkController.getInstance(marker.func_145831_w()).updateColor(marker, marker.func_174877_v());
    }

    public void removeStationMarker(TileEntityStationMarker marker){
        stations.remove(marker);
        NetworkController.getInstance(marker.func_145831_w()).updateColor((TileEntityStationMarker)null, marker.func_174877_v());
    }

    public Collection<RailWrapper> getStationRails(EntityMinecart cart, Pattern destinationRegex){
        Set<RailWrapper> rails = new HashSet<>();
        Set<String> validNames = new HashSet<>();
        for(TileEntityStationMarker station : stations) {
            if(station.isCartApplicable(cart, destinationRegex)) {
                rails.addAll(station.getNeighborRails());
                validNames.add(station.getStationName());
            }
        }

        //Make sure to include stations that don't match themselves, but other stations with the same name do.
        for(TileEntityStationMarker station : stations) {
            if(validNames.contains(station.getStationName())) {
                rails.addAll(station.getNeighborRails());
            }
        }
        return rails;
    }

    public Iterable<RailWrapper> getAllRails(){
        Iterable<RailWrapper> iterable = Collections.emptyList();
        for(Map<BlockPos, RailWrapper> map : railCache.values()) {
            iterable = Iterables.concat(iterable, map.values());
        }
        return iterable;
    }

    public static String[] getAllStationNames(){
        return getInstance(0).allStationNames;
    }

    public static void setAllStationNames(String[] stationNames){
        getInstance(0).allStationNames = stationNames;
    }

    public static void syncStationNames(EntityPlayerMP player){
        Set<String> stationNames = new HashSet<>();
        stationNames.add("ITEM");
        for(RailCacheManager manager : INSTANCES.values()) {
            for(TileEntityStationMarker station : manager.stations) {
                if(!station.getStationName().equals("")) stationNames.add(station.getStationName());
            }
        }
        String[] array = stationNames.toArray(new String[stationNames.size()]);
        Arrays.sort(array);
        NetworkHandler.sendTo(new PacketSyncStationNames(array), player);
    }
}
